@@ -3,7 +3,6 @@
 
 #pragma once
 
-#include "omittable.h"
 #include "util.h"
 
 #include <QtCore/QDate>
@@ -16,17 +15,33 @@
 
 #include <type_traits>
 #include <vector>
+#include <array>
 #include <variant>
+#include <optional>
 
 class QVariant;
 
 namespace Quotient {
+
+inline void editSubobject(QJsonObject& json, auto key, std::invocable<QJsonObject&> auto visitor)
+{
+    auto subObject = json.take(key).toObject();
+    visitor(subObject);
+    json.insert(key, subObject);
+}
+
+inline void replaceSubvalue(QJsonObject& json, auto topLevelKey, auto subKey, QJsonValue subValue)
+{
+    editSubobject(json, topLevelKey, [subKey, subValue](QJsonObject& innerJson) {
+        innerJson.insert(subKey, subValue);
+    });
+}
+
 template <typename T>
-struct JsonObjectConverter {
-    // To be implemented in specialisations
-    static void dumpTo(QJsonObject&, const T&) = delete;
-    static void fillFrom(const QJsonObject&, T&) = delete;
-};
+struct JsonObjectConverter;
+// Specialisations should implement either or both of:
+//static void dumpTo(QJsonObject&, const T&); // For toJson() and fillJson() to work
+//static void fillFrom(const QJsonObject&, T&); // For fromJson() and fillFromJson() to work
 
 template <typename PodT, typename JsonT>
 PodT fromJson(const JsonT&);
@@ -58,10 +73,6 @@ struct JsonObjectUnpacker {
 //! of overloading toJson() and specialising fromJson().
 template <typename T>
 struct JsonConverter : JsonObjectUnpacker<T> {
-    // Unfortunately, if constexpr doesn't work with dump() and T::toJson
-    // because trying to check invocability of T::toJson hits a hard
-    // (non-SFINAE) compilation error if the member is not there. Hence a bit
-    // more verbose SFINAE construct in _impl::JsonExporter.
     static auto dump(const T& data)
     {
         if constexpr (requires() { data.toJson(); })
@@ -136,9 +147,8 @@ inline void fillFromJson(const QJsonValue& jv, T& pod)
 }
 
 namespace _impl {
-    void warnUnknownEnumValue(const QString& stringValue,
-                              const char* enumTypeName);
-    void reportEnumOutOfBounds(uint32_t v, const char* enumTypeName);
+    QUOTIENT_API void warnUnknownEnumValue(const QString& stringValue, const char* enumTypeName);
+    QUOTIENT_API void reportEnumOutOfBounds(uint32_t v, const char* enumTypeName);
 }
 
 //! \brief Facility string-to-enum converter
@@ -150,13 +160,12 @@ namespace _impl {
 //! matching respective enum values, 0-based.
 //! \sa enumToJsonString
 template <typename EnumT, typename EnumStringValuesT>
-EnumT enumFromJsonString(const QString& s, const EnumStringValuesT& enumValues,
-                         EnumT defaultValue)
+inline EnumT enumFromJsonString(const QString& s, const EnumStringValuesT& enumValues,
+                                EnumT defaultValue)
 {
     static_assert(std::is_unsigned_v<std::underlying_type_t<EnumT>>);
-    if (const auto it = std::find(cbegin(enumValues), cend(enumValues), s);
-        it != cend(enumValues))
-        return EnumT(it - cbegin(enumValues));
+    if (const auto it = std::ranges::find(enumValues, s); it != cend(enumValues))
+        return static_cast<EnumT>(it - cbegin(enumValues));
 
     if (!s.isEmpty())
         _impl::warnUnknownEnumValue(s, qt_getEnumName(EnumT()));
@@ -173,7 +182,7 @@ EnumT enumFromJsonString(const QString& s, const EnumStringValuesT& enumValues,
 //!       }</tt> (mind the gap at value 0, in particular).
 //! \sa enumFromJsonString
 template <typename EnumT, typename EnumStringValuesT>
-QString enumToJsonString(EnumT v, const EnumStringValuesT& enumValues)
+inline QString enumToJsonString(EnumT v, const EnumStringValuesT& enumValues)
 {
     static_assert(std::is_unsigned_v<std::underlying_type_t<EnumT>>);
     if (v < size(enumValues))
@@ -196,14 +205,13 @@ QString enumToJsonString(EnumT v, const EnumStringValuesT& enumValues)
 //! \note This function does not support flag combinations.
 //! \sa QUO_DECLARE_FLAGS, QUO_DECLARE_FLAGS_NS
 template <typename FlagT, typename FlagStringValuesT>
-FlagT flagFromJsonString(const QString& s, const FlagStringValuesT& flagValues,
-                         FlagT defaultValue = FlagT(0U))
+inline FlagT flagFromJsonString(const QString& s, const FlagStringValuesT& flagValues,
+                                FlagT defaultValue = FlagT(0U))
 {
     // Enums based on signed integers don't make much sense for flag types
     static_assert(std::is_unsigned_v<std::underlying_type_t<FlagT>>);
-    if (const auto it = std::find(cbegin(flagValues), cend(flagValues), s);
-        it != cend(flagValues))
-        return FlagT(1U << (it - cbegin(flagValues)));
+    if (const auto it = std::ranges::find(flagValues, s); it != cend(flagValues))
+        return static_cast<FlagT>(1U << (it - cbegin(flagValues)));
 
     if (!s.isEmpty())
         _impl::warnUnknownEnumValue(s, qt_getEnumName(FlagT()));
@@ -211,25 +219,20 @@ FlagT flagFromJsonString(const QString& s, const FlagStringValuesT& flagValues,
 }
 
 template <typename FlagT, typename FlagStringValuesT>
-QString flagToJsonString(FlagT v, const FlagStringValuesT& flagValues)
+inline QString flagToJsonString(FlagT v, const FlagStringValuesT& flagValues)
 {
     static_assert(std::is_unsigned_v<std::underlying_type_t<FlagT>>);
-    if (const auto offset =
-            qCountTrailingZeroBits(std::underlying_type_t<FlagT>(v));
-        offset < size(flagValues)) //
-    {
+    if (const auto offset = std::countr_zero(std::to_underlying(v)); offset < ssize(flagValues))
         return flagValues[offset];
-    }
 
-    _impl::reportEnumOutOfBounds(static_cast<uint32_t>(v),
-                                 qt_getEnumName(FlagT()));
+    _impl::reportEnumOutOfBounds(static_cast<uint32_t>(v), qt_getEnumName(FlagT()));
     Q_ASSERT(false);
     return {};
 }
 
 // Specialisations
 
-template<>
+template <>
 inline bool fromJson(const QJsonValue& jv) { return jv.toBool(); }
 
 template <>
@@ -321,22 +324,22 @@ struct JsonConverter<std::variant<QString, T>> {
 };
 
 template <typename T>
-struct JsonConverter<Omittable<T>> {
-    static QJsonValue dump(const Omittable<T>& from)
+struct JsonConverter<std::optional<T>> {
+    static QJsonValue dump(const std::optional<T>& from)
     {
         return from.has_value() ? toJson(*from) : QJsonValue();
     }
-    static Omittable<T> load(const QJsonValue& jv)
+    static std::optional<T> load(const QJsonValue& jv)
     {
         if (jv.isUndefined() || jv.isNull())
-            return none;
+            return std::nullopt;
         return fromJson<T>(jv);
     }
 };
 
-template <typename VectorT, typename T = typename VectorT::value_type>
+template <typename ContT>
 struct JsonArrayConverter {
-    static auto dump(const VectorT& vals)
+    static auto dump(const ContT& vals)
     {
         QJsonArray ja;
         for (const auto& v : vals)
@@ -345,15 +348,15 @@ struct JsonArrayConverter {
     }
     static auto load(const QJsonArray& ja)
     {
-        VectorT vect;
-        vect.reserve(typename VectorT::size_type(ja.size()));
-        // NB: Make sure to pass QJsonValue to fromJson<> so that it could
-        // hit the correct overload and not fall back to the generic fromJson
-        // that treats everything as an object. See also the explanation in
-        // the commit introducing these lines.
-        for (const QJsonValue v : ja)
-            vect.push_back(fromJson<T>(v));
-        return vect;
+        ContT vals;
+        vals.reserve(static_cast<typename ContT::size_type>(ja.size()));
+        // NB: Make sure fromJson<> gets QJsonValue (not QJsonValue*Ref)
+        // to avoid it falling back to the generic implementation that treats
+        // everything as an object. See also the message of commit 20f01303b
+        // that introduced these lines.
+        for (const auto& v : ja)
+            vals.push_back(fromJson<typename ContT::value_type, QJsonValue>(v));
+        return vals;
     }
     static auto load(const QJsonValue& jv) { return load(jv.toArray()); }
     static auto load(const QJsonDocument& jd) { return load(jd.array()); }
@@ -363,10 +366,34 @@ template <typename T>
 struct JsonConverter<std::vector<T>>
     : public JsonArrayConverter<std::vector<T>> {};
 
-#if QT_VERSION_MAJOR < 6 // QVector is an alias of QList in Qt6 but not in Qt 5
-template <typename T>
-struct JsonConverter<QVector<T>> : public JsonArrayConverter<QVector<T>> {};
-#endif
+template <typename T, size_t N>
+struct JsonConverter<std::array<T, N>> {
+    // The size of std::array is known at compile-time and those arrays
+    // are usually short. The common conversion logic therefore is to expand
+    // the passed source array into a pack of values converted with to/fromJson
+    // and then construct the target array list-initialised with that pack.
+    // For load(), this implies that if QJsonArray is not of the right size,
+    // the resulting std::array will not have extra values or will have empty
+    // values at the end - silently.
+    static constexpr std::make_index_sequence<N> Indices{};
+    template <typename TargetT, size_t... I>
+    static auto staticTransform(const auto& source, std::index_sequence<I...>,
+                                auto unaryFn)
+    {
+        return TargetT { unaryFn(source[I])... };
+    }
+    static auto dump(const std::array<T, N> a)
+    {
+        return staticTransform<QJsonArray>(a, Indices, [](const T& v) {
+            return toJson(v);
+        });
+    }
+    static auto load(const QJsonArray& ja)
+    {
+        return staticTransform<std::array<T, N>>(ja, Indices,
+                                                 fromJson<T, QJsonValue>);
+    }
+};
 
 template <typename T>
 struct JsonConverter<QList<T>> : public JsonArrayConverter<QList<T>> {};
@@ -404,11 +431,11 @@ struct HashMapFromJson {
     static void fillFrom(const QJsonObject& jo, HashMapT& h)
     {
         h.reserve(h.size() + jo.size());
-        // NB: the QJsonValue cast below is for the same reason as in
-        // JsonArrayConverter
+        // NB: coercing the passed value to QJsonValue below is for
+        // the same reason as in JsonArrayConverter
         for (auto it = jo.begin(); it != jo.end(); ++it)
-            h[it.key()] = fromJson<typename HashMapT::mapped_type>(
-                QJsonValue(it.value()));
+            h[it.key()] = fromJson<typename HashMapT::mapped_type, QJsonValue>(
+                it.value());
     }
 };
 
@@ -429,28 +456,28 @@ QVariantHash QUOTIENT_API fromJson(const QJsonValue& jv);
 constexpr bool IfNotEmpty = false;
 
 namespace _impl {
-    template <typename ValT>
-    inline void addTo(QJsonObject& o, const QString& k, ValT&& v)
+    template <typename KeyT, typename ValT>
+    inline void addTo(QJsonObject& o, KeyT&& k, ValT&& v)
     {
-        o.insert(k, toJson(v));
+        o.insert(std::forward<KeyT>(k), toJson(std::forward<ValT>(v)));
     }
 
-    template <typename ValT>
-    inline void addTo(QUrlQuery& q, const QString& k, ValT&& v)
+    inline void addTo(QUrlQuery& q, const QString& k, auto v)
+        requires requires { u"%1"_s.arg(v); }
     {
-        q.addQueryItem(k, QStringLiteral("%1").arg(v));
+        q.addQueryItem(k, u"%1"_s.arg(v));
     }
 
     // OpenAPI is entirely JSON-based, which means representing bools as
     // textual true/false, rather than 1/0.
     inline void addTo(QUrlQuery& q, const QString& k, bool v)
     {
-        q.addQueryItem(k, v ? QStringLiteral("true") : QStringLiteral("false"));
+        q.addQueryItem(k, v ? u"true"_s : u"false"_s);
     }
 
     inline void addTo(QUrlQuery& q, const QString& k, const QUrl& v)
     {
-        q.addQueryItem(k, v.toEncoded());
+        q.addQueryItem(k, QString::fromLatin1(v.toEncoded()));
     }
 
     inline void addTo(QUrlQuery& q, const QString& k, const QStringList& vals)
@@ -459,39 +486,44 @@ namespace _impl {
             q.addQueryItem(k, v);
     }
 
+    template <typename ValT>
+    inline void addTo(QUrlQuery& q, const QString&, const QHash<QString, ValT>& fields)
+    {
+        for (const auto& [k, v] : fields.asKeyValueRange())
+            addTo(q, k, v);
+    }
+
     // This one is for types that don't have isEmpty() and for all types
     // when Force is true
-    template <typename ValT, bool Force = true, typename = bool>
+    template <typename ValT, bool Force = true>
     struct AddNode {
-        template <typename ContT, typename ForwardedT>
-        static void impl(ContT& container, const QString& key,
-                         ForwardedT&& value)
+        template <typename KeyT, typename ForwardedT>
+        static void impl(auto& container, KeyT&& key, ForwardedT&& value)
         {
-            addTo(container, key, std::forward<ForwardedT>(value));
+            addTo(container, std::forward<KeyT>(key), std::forward<ForwardedT>(value));
         }
     };
 
     // This one is for types that have isEmpty() when Force is false
     template <typename ValT>
-    struct AddNode<ValT, IfNotEmpty, decltype(std::declval<ValT>().isEmpty())> {
-        template <typename ContT, typename ForwardedT>
-        static void impl(ContT& container, const QString& key,
-                         ForwardedT&& value)
+        requires requires(ValT v) { v.isEmpty(); }
+    struct AddNode<ValT, IfNotEmpty> {
+        template <typename KeyT, typename ForwardedT>
+        static void impl(auto& container, KeyT&& key, ForwardedT&& value)
         {
             if (!value.isEmpty())
-                addTo(container, key, std::forward<ForwardedT>(value));
+                addTo(container, std::forward<KeyT>(key), std::forward<ForwardedT>(value));
         }
     };
 
-    // This one unfolds Omittable<> (also only when IfNotEmpty is requested)
+    // This one unfolds optionals (also only when IfNotEmpty is requested)
     template <typename ValT>
-    struct AddNode<Omittable<ValT>, IfNotEmpty> {
-        template <typename ContT, typename OmittableT>
-        static void impl(ContT& container, const QString& key,
-                         const OmittableT& value)
+    struct AddNode<std::optional<ValT>, IfNotEmpty> {
+        template <typename KeyT>
+        static void impl(auto& container, KeyT&& key, const auto& optValue)
         {
-            if (value)
-                addTo(container, key, *value);
+            if (optValue)
+                addTo(container, std::forward<KeyT>(key), *optValue);
         }
     };
 } // namespace _impl
@@ -503,7 +535,7 @@ namespace _impl {
  * template parameter) taking into account the value "emptiness".
  * With IfNotEmpty, \p value is NOT added to the container if and only if:
  * - it has a method `isEmpty()` and `value.isEmpty() == true`, or
- * - it's an `Omittable<>` and `value.omitted() == true`.
+ * - it's an optional that has no value (`nullopt`).
  *
  * If \p container is a QUrlQuery, an attempt to fit \p value into it is
  * made as follows:
@@ -519,23 +551,23 @@ namespace _impl {
  *               by default; passing IfNotEmpty or false for this parameter
  *               enables emptiness checks as described above
  */
-template <bool Force = true, typename ContT, typename ValT>
-inline void addParam(ContT& container, const QString& key, ValT&& value)
+template <bool Force = true, typename ContT, typename KeyT, typename ValT>
+inline void addParam(ContT& container, KeyT&& key, ValT&& value)
 {
-    _impl::AddNode<std::decay_t<ValT>, Force>::impl(container, key,
+    _impl::AddNode<std::decay_t<ValT>, Force>::impl(container, std::forward<KeyT>(key),
                                                     std::forward<ValT>(value));
 }
 
 // This is a facility function to convert camelCase method/variable names
 // used throughout Quotient to snake_case JSON keys - see usage in
-// single_key_value.h and event.h (DEFINE_CONTENT_GETTER macro).
+// single_key_value.h and event.h (QUO_CONTENT_GETTER macro).
 inline auto toSnakeCase(QLatin1String s)
 {
     QString result { s };
     for (auto it = result.begin(); it != result.end(); ++it)
         if (it->isUpper()) {
             const auto offset = static_cast<int>(it - result.begin());
-            result.insert(offset, '_'); // NB: invalidates iterators
+            result.insert(offset, u'_'); // NB: invalidates iterators
             it = result.begin() + offset + 1;
             *it = it->toLower();
         }
